"""Module that defines AttributeModel."""

import inspect
import logging
import pprint
from typing import Callable, Dict

from qtpy import QtGui

from prettyqt import constants, custom_models, gui
from prettyqt.utils import treeitem


logger = logging.getLogger(__name__)


_PRETTY_PRINTER = pprint.PrettyPrinter(indent=4)

_ALL_PREDICATES = (
    inspect.isgenerator,
    inspect.istraceback,
    inspect.isframe,
    inspect.iscode,
    inspect.isabstract,
    inspect.isgetsetdescriptor,
    inspect.ismemberdescriptor,
)

CALLABLE_COLOR = QtGui.QBrush(gui.Color("mediumblue"))
REGULAR_COLOR = QtGui.QBrush(gui.Color("black"))

REGULAR_FONT = QtGui.QFont()  # Font for members (non-functions)
SPECIAL_ATTR_FONT = QtGui.QFont()  # Font for __special_attributes__
SPECIAL_ATTR_FONT.setItalic(True)


TYPE_CHECK: Dict[Callable, str] = {
    inspect.isasyncgen: "Async generator",
    inspect.ismethoddescriptor: "Method descriptor",
    inspect.ismethod: "Method",
    inspect.isfunction: "Function",
    inspect.isgeneratorfunction: "Generator function",
    inspect.isclass: "Class",
    inspect.ismodule: "Module",
    inspect.isdatadescriptor: "Data descriptor",
    inspect.isbuiltin: "Builtin",
}


def get_type(tree_item: treeitem.TreeItem) -> str:

    for k, v in TYPE_CHECK.items():
        if k(tree_item.obj):
            return v
    return "Attribute"


def safe_data_fn(
    obj_fn: Callable,
    log_exceptions: bool = False,
) -> Callable:
    """Create a function that returns an empty string in case of an exception.

    :param fnobj_fn: function that will be wrapped
    :type obj_fn: object to basestring function
    :returns: function that can be used as AttributeModel data_fn attribute
    :rtype: custom_models.treeitem.TreeItem to string function
    """

    def data_fn(tree_item: treeitem.TreeItem) -> str:
        """Call the obj_fn(tree_item.obj).

        Returns empty string in case of an error
        """
        try:
            return str(obj_fn(tree_item.obj))
        except Exception as ex:
            if log_exceptions:
                logger.exception(ex)
            return ""

    return data_fn


#######################
# Column definitions ##
#######################

ATTR_MODEL_NAME = custom_models.ColumnItem(
    name="Name",
    doc="The name of the object.",
    label=lambda tree_item: tree_item.obj_name if tree_item.obj_name else "<root>",
    col_visible=True,
    width="small",
    font=lambda x: SPECIAL_ATTR_FONT if x.is_special_attribute else REGULAR_FONT,
    foreground_color=lambda x: CALLABLE_COLOR if callable(x.obj) else REGULAR_COLOR,
)

ATTR_MODEL_DESCRIPTION = custom_models.ColumnItem(
    name="Description",
    doc="Description of the object.",
    label=lambda tree_item: get_type(tree_item),
    col_visible=True,
    width="small",
)

ATTR_MODEL_PATH = custom_models.ColumnItem(
    name="Path",
    doc="A path to the data: e.g. var[1]['a'].item",
    label=lambda tree_item: tree_item.obj_path if tree_item.obj_path else "<root>",
    col_visible=True,
    width="medium",
)


ATTR_MODEL_STR = custom_models.ColumnItem(
    name="str",
    doc="""The string representation of the object using the str() function.""",
    label=lambda tree_item: str(tree_item.obj),
    col_visible=False,
    width="medium",
    line_wrap="boundary_or_anywhere",
)

ATTR_MODEL_REPR = custom_models.ColumnItem(
    name="repr",
    doc="The string representation of the object using the repr() function.",
    label=lambda tree_item: repr(tree_item.obj),
    col_visible=True,
    width="medium",
    line_wrap="boundary_or_anywhere",
)

ATTR_MODEL_TYPE = custom_models.ColumnItem(
    name="Type",
    doc="Type of the object determined using the builtin type() function",
    label=lambda tree_item: str(type(tree_item.obj)),
    col_visible=False,
    width="medium",
)

ATTR_MODEL_CLASS = custom_models.ColumnItem(
    name="Type name",
    doc="The name of the class of the object via obj.__class__.__name__",
    label=lambda tree_item: type(tree_item.obj).__name__,
    col_visible=True,
    width="medium",
)

ATTR_MODEL_LENGTH = custom_models.ColumnItem(
    name="Length",
    doc="The length of the object using the len() function",
    # data_fn     = tio_length,
    label=safe_data_fn(len),
    col_visible=False,
    alignment=constants.ALIGN_RIGHT,
    width="small",
)

ATTR_MODEL_ID = custom_models.ColumnItem(
    name="id",
    doc="The identifier of the object with calculated using the id() function",
    label=lambda tree_item: f"0x{id(tree_item.obj):X}",
    col_visible=False,
    alignment=constants.ALIGN_RIGHT,
    width="small",
)

ATTR_MODEL_IS_ATTRIBUTE = custom_models.ColumnItem(
    name="Attribute",
    doc="""The object is an attribute of the parent (opposed to e.g. a list element).
                     Attributes are displayed in italics in the table.
                  """,
    label=lambda x: "" if x.is_attribute is None else str(x.is_attribute),
    col_visible=False,
    width="small",
)

ATTR_MODEL_CALLABLE = custom_models.ColumnItem(
    name="Callable",
    doc="""True if the object is callable.
                     Determined with the `callable` built-in function.
                     Callable objects are displayed in blue in the table.
                  """,
    label=None,
    checkstate=lambda tree_item: callable(tree_item.obj),
    col_visible=True,
    width="small",
)

ATTR_MODEL_IS_ROUTINE = custom_models.ColumnItem(
    name="Routine",
    doc="""True if the object is a user-defined or built-in function or method.
                     Determined with the inspect.isroutine() method.
                  """,
    label=None,
    checkstate=lambda tree_item: inspect.isroutine(tree_item.obj),
    col_visible=False,
    width="small",
)

ATTR_MODEL_IS_BUILTIN = custom_models.ColumnItem(
    name="Builtin",
    doc="""True if the object is a builtin.
                     Determined with the inspect.isbuiltin() method.
                  """,
    label=None,
    checkstate=lambda tree_item: inspect.isbuiltin(tree_item.obj),
    col_visible=False,
    width="small",
)

ATTR_MODEL_PRED = custom_models.ColumnItem(
    name="Inspect predicates",
    doc="Predicates from the inspect module",
    label=lambda x: ", ".join(pred.__name__ for pred in _ALL_PREDICATES if pred(x.obj)),
    col_visible=False,
    width="medium",
)

ATTR_MODEL_PRETTY_PRINT = custom_models.ColumnItem(
    name="Pretty print",
    doc="Pretty printed representation of the object using the pprint module.",
    label=lambda tree_item: _PRETTY_PRINTER.pformat(tree_item.obj),
    col_visible=False,
    width="medium",
)


ATTR_MODEL_GET_DOC = custom_models.ColumnItem(
    name="inspect.getdoc",
    doc="The object's doc string, leaned up by inspect.getdoc()",
    label=safe_data_fn(inspect.getdoc),
    col_visible=False,
    width="medium",
)

ATTR_MODEL_GET_COMMENTS = custom_models.ColumnItem(
    name="inspect.getcomments",
    doc="Comments above the object's definition. Retrieved using inspect.getcomments()",
    label=lambda tree_item: inspect.getcomments(tree_item.obj),
    col_visible=False,
    width="medium",
)

ATTR_MODEL_GET_MODULE = custom_models.ColumnItem(
    name="inspect.getmodule",
    doc="The object's module. Retrieved using inspect.module",
    label=safe_data_fn(inspect.getmodule),
    col_visible=False,
    width="medium",
)

ATTR_MODEL_GET_FILE = custom_models.ColumnItem(
    name="inspect.getfile",
    doc="The object's file. Retrieved using inspect.getfile",
    label=safe_data_fn(inspect.getfile),
    col_visible=False,
    width="medium",
)

ATTR_MODEL_GET_SOURCE_FILE = custom_models.ColumnItem(
    name="inspect.getsourcefile",  # calls inspect.getfile()
    doc="The object's file. Retrieved using inspect.getsourcefile",
    label=safe_data_fn(inspect.getsourcefile),
    col_visible=False,
    width="medium",
)


ATTR_MODEL_GET_SOURCE = custom_models.ColumnItem(
    name="Inspect.getsource",
    doc="The source code of an object retrieved using inspect.getsource",
    label=safe_data_fn(inspect.getsource),
    col_visible=False,
    width="medium",
)


DEFAULT_ATTR_COLS = [
    ATTR_MODEL_NAME,
    ATTR_MODEL_DESCRIPTION,
    ATTR_MODEL_PATH,
    ATTR_MODEL_STR,
    ATTR_MODEL_REPR,
    ATTR_MODEL_LENGTH,
    ATTR_MODEL_TYPE,
    ATTR_MODEL_CLASS,
    ATTR_MODEL_ID,
    ATTR_MODEL_IS_ATTRIBUTE,
    ATTR_MODEL_CALLABLE,
    ATTR_MODEL_IS_ROUTINE,
    ATTR_MODEL_IS_BUILTIN,
    ATTR_MODEL_PRED,
    ATTR_MODEL_GET_MODULE,
    ATTR_MODEL_GET_FILE,
    ATTR_MODEL_GET_SOURCE_FILE,
]

DEFAULT_ATTR_DETAILS = [
    ATTR_MODEL_PATH,  # to allow for copy/paste
    ATTR_MODEL_STR,  # Too similar to unicode column
    ATTR_MODEL_REPR,
    ATTR_MODEL_PRETTY_PRINT,
    ATTR_MODEL_GET_DOC,
    ATTR_MODEL_GET_COMMENTS,
    # ATTR_MODEL_GET_MODULE, # not used, already in table
    ATTR_MODEL_GET_FILE,
    # ATTR_MODEL_GET_SOURCE_FILE,  # not used, already in table
    ATTR_MODEL_GET_SOURCE,
]

# Sanity check for duplicates
assert len(DEFAULT_ATTR_COLS) == len(set(DEFAULT_ATTR_COLS))
assert len(DEFAULT_ATTR_DETAILS) == len(set(DEFAULT_ATTR_DETAILS))



from typing import Literal

from qtpy import QtWidgets

from prettyqt import core
from prettyqt.utils import bidict


STANDARD_PIXMAP = dict(
    titlebar_min_button=QtWidgets.QStyle.SP_TitleBarMinButton,
    titlebar_menu_button=QtWidgets.QStyle.SP_TitleBarMenuButton,
    titlebar_max_button=QtWidgets.QStyle.SP_TitleBarMaxButton,
    titlebar_close_button=QtWidgets.QStyle.SP_TitleBarCloseButton,
    titlebar_normal_button=QtWidgets.QStyle.SP_TitleBarNormalButton,
    titlebar_shade_button=QtWidgets.QStyle.SP_TitleBarShadeButton,
    titlebar_unshade_button=QtWidgets.QStyle.SP_TitleBarUnshadeButton,
    titlebar_context_help_button=QtWidgets.QStyle.SP_TitleBarContextHelpButton,
    information=QtWidgets.QStyle.SP_MessageBoxInformation,
    warning=QtWidgets.QStyle.SP_MessageBoxWarning,
    critical=QtWidgets.QStyle.SP_MessageBoxCritical,
    question=QtWidgets.QStyle.SP_MessageBoxQuestion,
    desktop=QtWidgets.QStyle.SP_DesktopIcon,
    trash=QtWidgets.QStyle.SP_TrashIcon,
    computer=QtWidgets.QStyle.SP_ComputerIcon,
    drive_fd=QtWidgets.QStyle.SP_DriveFDIcon,
    drive_hd=QtWidgets.QStyle.SP_DriveHDIcon,
    drive_cd=QtWidgets.QStyle.SP_DriveCDIcon,
    drive_dvd=QtWidgets.QStyle.SP_DriveDVDIcon,
    drive_net=QtWidgets.QStyle.SP_DriveNetIcon,
    dir_home=QtWidgets.QStyle.SP_DirHomeIcon,
    dir_open=QtWidgets.QStyle.SP_DirOpenIcon,
    dir_closed=QtWidgets.QStyle.SP_DirClosedIcon,
    dir=QtWidgets.QStyle.SP_DirIcon,
    dir_link=QtWidgets.QStyle.SP_DirLinkIcon,
    dir_link_open=QtWidgets.QStyle.SP_DirLinkOpenIcon,
    file=QtWidgets.QStyle.SP_FileIcon,
    file_link=QtWidgets.QStyle.SP_FileLinkIcon,
    file_dialog_start=QtWidgets.QStyle.SP_FileDialogStart,
    file_dialog_end=QtWidgets.QStyle.SP_FileDialogEnd,
    file_dialog_to_parent=QtWidgets.QStyle.SP_FileDialogToParent,
    file_dialog_new_folder=QtWidgets.QStyle.SP_FileDialogNewFolder,
    file_dialog_detailed_view=QtWidgets.QStyle.SP_FileDialogDetailedView,
    file_dialog_info_view=QtWidgets.QStyle.SP_FileDialogInfoView,
    file_dialog_contents_view=QtWidgets.QStyle.SP_FileDialogContentsView,
    file_dialog_list_view=QtWidgets.QStyle.SP_FileDialogListView,
    file_dialog_back=QtWidgets.QStyle.SP_FileDialogBack,
    dockwidget_close_button=QtWidgets.QStyle.SP_DockWidgetCloseButton,
    toolbar_horizontal_extension=QtWidgets.QStyle.SP_ToolBarHorizontalExtensionButton,
    toolbar_vertical_extension=QtWidgets.QStyle.SP_ToolBarVerticalExtensionButton,
    dialog_ok=QtWidgets.QStyle.SP_DialogOkButton,
    dialog_cancel=QtWidgets.QStyle.SP_DialogCancelButton,
    dialog_help=QtWidgets.QStyle.SP_DialogHelpButton,
    dialog_open=QtWidgets.QStyle.SP_DialogOpenButton,
    dialog_save=QtWidgets.QStyle.SP_DialogSaveButton,
    dialog_close=QtWidgets.QStyle.SP_DialogCloseButton,
    dialog_apply=QtWidgets.QStyle.SP_DialogApplyButton,
    dialor_reset=QtWidgets.QStyle.SP_DialogResetButton,
    dialog_discard=QtWidgets.QStyle.SP_DialogDiscardButton,
    dialog_yes=QtWidgets.QStyle.SP_DialogYesButton,
    dialog_no=QtWidgets.QStyle.SP_DialogNoButton,
    arrow_up=QtWidgets.QStyle.SP_ArrowUp,
    arrow_down=QtWidgets.QStyle.SP_ArrowDown,
    arrow_left=QtWidgets.QStyle.SP_ArrowLeft,
    arrow_right=QtWidgets.QStyle.SP_ArrowRight,
    arrow_back=QtWidgets.QStyle.SP_ArrowBack,
    arrow_forward=QtWidgets.QStyle.SP_ArrowForward,
    command_link=QtWidgets.QStyle.SP_CommandLink,
    vista_shield=QtWidgets.QStyle.SP_VistaShield,
    browser_reload=QtWidgets.QStyle.SP_BrowserReload,
    browser_stop=QtWidgets.QStyle.SP_BrowserStop,
    media_play=QtWidgets.QStyle.SP_MediaPlay,
    media_stop=QtWidgets.QStyle.SP_MediaStop,
    media_pause=QtWidgets.QStyle.SP_MediaPause,
    media_skip_forward=QtWidgets.QStyle.SP_MediaSkipForward,
    media_skip_backward=QtWidgets.QStyle.SP_MediaSkipBackward,
    media_seek_forward=QtWidgets.QStyle.SP_MediaSeekForward,
    media_seek_backward=QtWidgets.QStyle.SP_MediaSeekBackward,
    media_volume=QtWidgets.QStyle.SP_MediaVolume,
    media_volume_muted=QtWidgets.QStyle.SP_MediaVolumeMuted,
    lineedit_clear=QtWidgets.QStyle.SP_LineEditClearButton,
    custom_base=QtWidgets.QStyle.SP_CustomBase,
)

StandardPixmapStr = Literal[
    "titlebar_min_button=",
    "titlebar_menu_button",
    "titlebar_max_button",
    "titlebar_close_button",
    "titlebar_normal_button",
    "titlebar_shade_button",
    "titlebar_unshade_button",
    "titlebar_context_help_button",
    "information",
    "warning",
    "critical",
    "question",
    "desktop",
    "trash",
    "computer",
    "drive_fd",
    "drive_hd",
    "drive_cd",
    "drive_dvd",
    "drive_net",
    "dir_home",
    "dir_open",
    "dir_closed",
    "dir",
    "dir_link",
    "dir_link_open",
    "file",
    "file_link",
    "file_dialog_start",
    "file_dialog_end",
    "file_dialog_to_parent",
    "file_dialog_new_folder",
    "file_dialog_detailed_view",
    "file_dialog_info_view",
    "file_dialog_contents_view",
    "file_dialog_list_view",
    "file_dialog_back",
    "dockwidget_close_button",
    "toolbar_horizontal_extension",
    "toolbar_vertical_extension",
    "dialog_ok",
    "dialog_cancel",
    "dialog_help",
    "dialog_open",
    "dialog_save",
    "dialog_close",
    "dialog_apply",
    "dialor_reset",
    "dialog_discard",
    "dialog_yes",
    "dialog_no",
    "arrow_up",
    "arrow_down",
    "arrow_left",
    "arrow_right",
    "arrow_back",
    "arrow_forward",
    "command_link",
    "vista_shield",
    "browser_reload",
    "browser_stop",
    "media_play",
    "media_stop",
    "media_pause",
    "media_skip_forward",
    "media_skip_backward",
    "media_seek_forward",
    "media_seek_backward",
    "media_volume",
    "media_volume_muted",
    "lineedit_clear",
    "custom_base",
]

if core.VersionNumber.get_qt_version() >= (5, 14, 0):
    STANDARD_PIXMAP["dialog_yes_to_all"] = QtWidgets.QStyle.SP_DialogYesToAllButton
    STANDARD_PIXMAP["dialog_no_to_all"] = QtWidgets.QStyle.SP_DialogNoToAllButton
    STANDARD_PIXMAP["dialog_save_all"] = QtWidgets.QStyle.SP_DialogSaveAllButton
    STANDARD_PIXMAP["dialog_abort"] = QtWidgets.QStyle.SP_DialogAbortButton
    STANDARD_PIXMAP["dialog_retry"] = QtWidgets.QStyle.SP_DialogRetryButton
    STANDARD_PIXMAP["dialog_ignore"] = QtWidgets.QStyle.SP_DialogIgnoreButton
    STANDARD_PIXMAP["restore_defaults"] = QtWidgets.QStyle.SP_RestoreDefaultsButton

COMPLEX_CONTROL = bidict(
    spinbox=QtWidgets.QStyle.CC_SpinBox,
    combobox=QtWidgets.QStyle.CC_ComboBox,
    scrollbar=QtWidgets.QStyle.CC_ScrollBar,
    slider=QtWidgets.QStyle.CC_Slider,
    toolbutton=QtWidgets.QStyle.CC_ToolButton,
    titlebar=QtWidgets.QStyle.CC_TitleBar,
    groupbox=QtWidgets.QStyle.CC_GroupBox,
    dial=QtWidgets.QStyle.CC_Dial,
    mdi_controls=QtWidgets.QStyle.CC_MdiControls,
    custom_base=QtWidgets.QStyle.CC_CustomBase,
)

ComplexControlStr = Literal[
    "spinbox",
    "combobox",
    "scrollbar",
    "slider",
    "toolbutton",
    "titlebar",
    "groupbox",
    "dial",
    "mdi_controls",
    "custom_base",
]

SIMPLE_CONTROLS = dict(
    none=QtWidgets.QStyle.SC_None,
    scrollbar_add_line=QtWidgets.QStyle.SC_ScrollBarAddLine,
    scrollbar_sub_line=QtWidgets.QStyle.SC_ScrollBarSubLine,
    scrollbar_add_page=QtWidgets.QStyle.SC_ScrollBarAddPage,
    scrollbar_sub_page=QtWidgets.QStyle.SC_ScrollBarSubPage,
    scrollbar_first=QtWidgets.QStyle.SC_ScrollBarFirst,
    scrollbar_last=QtWidgets.QStyle.SC_ScrollBarLast,
    scrollbar_slider=QtWidgets.QStyle.SC_ScrollBarSlider,
    scrollbar_groove=QtWidgets.QStyle.SC_ScrollBarGroove,
    spinbox_up=QtWidgets.QStyle.SC_SpinBoxUp,
    spinbox_down=QtWidgets.QStyle.SC_SpinBoxDown,
    spinbox_frame=QtWidgets.QStyle.SC_SpinBoxFrame,
    spinbox_edit_field=QtWidgets.QStyle.SC_SpinBoxEditField,
    combobox_edit_field=QtWidgets.QStyle.SC_ComboBoxEditField,
    combobox_arrow=QtWidgets.QStyle.SC_ComboBoxArrow,
    combobox_frame=QtWidgets.QStyle.SC_ComboBoxFrame,
    combobox_list_box_popup=QtWidgets.QStyle.SC_ComboBoxListBoxPopup,
    slider_groove=QtWidgets.QStyle.SC_SliderGroove,
    slider_handle=QtWidgets.QStyle.SC_SliderHandle,
    slider_tickmarks=QtWidgets.QStyle.SC_SliderTickmarks,
    toolbutton=QtWidgets.QStyle.SC_ToolButton,
    toolbutton_menu=QtWidgets.QStyle.SC_ToolButtonMenu,
    titlebar_sys_menu=QtWidgets.QStyle.SC_TitleBarSysMenu,
    titlebar_min_button=QtWidgets.QStyle.SC_TitleBarMinButton,
    titlebar_max_button=QtWidgets.QStyle.SC_TitleBarMaxButton,
    titlebar_close_button=QtWidgets.QStyle.SC_TitleBarCloseButton,
    titlebar_label=QtWidgets.QStyle.SC_TitleBarLabel,
    titlebar_normal_button=QtWidgets.QStyle.SC_TitleBarNormalButton,
    titlebar_shade_button=QtWidgets.QStyle.SC_TitleBarShadeButton,
    titlebar_unshade_button=QtWidgets.QStyle.SC_TitleBarUnshadeButton,
    titlebar_context_help_button=QtWidgets.QStyle.SC_TitleBarContextHelpButton,
    dial_handle=QtWidgets.QStyle.SC_DialHandle,
    dial_groove=QtWidgets.QStyle.SC_DialGroove,
    dial_tickmarks=QtWidgets.QStyle.SC_DialTickmarks,
    groupbox_frame=QtWidgets.QStyle.SC_GroupBoxFrame,
    groupbox_label=QtWidgets.QStyle.SC_GroupBoxLabel,
    groupbox_checkbox=QtWidgets.QStyle.SC_GroupBoxCheckBox,
    groupbox_contents=QtWidgets.QStyle.SC_GroupBoxContents,
    mdi_normal_button=QtWidgets.QStyle.SC_MdiNormalButton,
    mdi_min_button=QtWidgets.QStyle.SC_MdiMinButton,
    mdi_close_button=QtWidgets.QStyle.SC_MdiCloseButton,
    all=QtWidgets.QStyle.SC_All,
)


class Style(QtWidgets.QStyle):
    pass



"""A lightweight module handling iconic fonts.

It is designed to provide a simple way for creating QtGui.QIcons from glyphs.

From a user's viewpoint, the main entry point is the ``IconicFont`` class which
contains methods for loading new iconic fonts with their character map and
methods returning instances of ``QIcon``.

"""

# Standard library imports
import hashlib
import json
import pathlib
from typing import Any, Dict, Iterable, List, Optional, Tuple
import warnings

from qtpy import QtCore, QtGui

from prettyqt import constants, core, gui


# Third party imports


# Linux packagers, please set this to True if you want to make qtawesome
# use system fonts
SYSTEM_FONTS = False

# MD5 Hashes for font files bundled with qtawesome:
MD5_HASHES = {
    "fontawesome4.7-webfont.ttf": "b06871f281fee6b241d60582ae9369b9",
    "fontawesome5-regular-webfont.ttf": "6a745dc6a0871f350b0219f5a2678838",
    "fontawesome5-solid-webfont.ttf": "acf50f59802f20d8b45220eaae532a1c",
    "fontawesome5-brands-webfont.ttf": "ed2b8bf117160466ba6220a8f1da54a4",
    "elusiveicons-webfont.ttf": "207966b04c032d5b873fd595a211582e",
    "materialdesignicons-webfont.ttf": "b0fd91bb29dcb296a9a37f8bda0a2d85",
}

_default_options = {
    "color": gui.Color(50, 50, 50),
    "color_disabled": gui.Color(150, 150, 150),
    "opacity": 1.0,
    "scale_factor": 1.0,
}

ZERO_COORD = core.Point(0, 0)


def set_global_defaults(**kwargs):
    """Set global defaults for the options passed to the icon painter."""
    valid_options = [
        "active",
        "selected",
        "disabled",
        "on",
        "off",
        "on_active",
        "on_selected",
        "on_disabled",
        "off_active",
        "off_selected",
        "off_disabled",
        "color",
        "color_on",
        "color_off",
        "color_active",
        "color_selected",
        "color_disabled",
        "color_on_selected",
        "color_on_active",
        "color_on_disabled",
        "color_off_selected",
        "color_off_active",
        "color_off_disabled",
        "animation",
        "offset",
        "scale_factor",
        "rotated",
        "hflip",
        "vflip",
    ]
    for kw in kwargs:
        if kw not in valid_options:
            raise KeyError(f"Invalid option '{kw}'")
        _default_options[kw] = kwargs[kw]


class CharIconPainter:
    """Char icon painter."""

    def paint(self, iconic, painter, rect, mode, state, options: List[Dict[str, Any]]):
        for opt in options:
            self._paint_icon(iconic, painter, rect, mode, state, opt)

    def _paint_icon(self, iconic, painter, rect, mode, state, options: Dict[str, Any]):
        color = options["color"]
        char = options["char"]

        color_options = {
            QtGui.QIcon.On: {
                QtGui.QIcon.Normal: (options["color_on"], options["on"]),
                QtGui.QIcon.Disabled: (
                    options["color_on_disabled"],
                    options["on_disabled"],
                ),
                QtGui.QIcon.Active: (options["color_on_active"], options["on_active"]),
                QtGui.QIcon.Selected: (
                    options["color_on_selected"],
                    options["on_selected"],
                ),
            },
            QtGui.QIcon.Off: {
                QtGui.QIcon.Normal: (options["color_off"], options["off"]),
                QtGui.QIcon.Disabled: (
                    options["color_off_disabled"],
                    options["off_disabled"],
                ),
                QtGui.QIcon.Active: (options["color_off_active"], options["off_active"]),
                QtGui.QIcon.Selected: (
                    options["color_off_selected"],
                    options["off_selected"],
                ),
            },
        }

        color, char = color_options[state][mode]

        painter.save()
        painter.setPen(gui.Color(color))

        # A 16 pixel-high icon yields a font size of 14, which is pixel perfect
        # for font-awesome. 16 * 0.875 = 14
        # The reason why the glyph size is smaller than the icon size is to
        # account for font bearing.

        draw_size = round(0.875 * rect.height() * options["scale_factor"])
        prefix = options["prefix"]

        # Animation setup hook
        animation = options.get("animation")
        if animation is not None:
            animation.setup(self, painter, rect)

        painter.setFont(iconic.font(prefix, draw_size))
        if "offset" in options:
            rect = core.Rect(rect)
            rect.translate(
                round(options["offset"][0] * rect.width()),
                round(options["offset"][1] * rect.height()),
            )

        if "vflip" in options and options["vflip"] is True:
            x_center = rect.width() * 0.5
            y_center = rect.height() * 0.5
            painter.translate(x_center, y_center)
            transfrom = gui.Transform()
            transfrom.scale(1, -1)
            painter.setTransform(transfrom, True)
            painter.translate(-x_center, -y_center)

        if "hflip" in options and options["hflip"] is True:
            x_center = rect.width() * 0.5
            y_center = rect.height() * 0.5
            painter.translate(x_center, y_center)
            transfrom = gui.Transform()
            transfrom.scale(-1, 1)
            painter.setTransform(transfrom, True)
            painter.translate(-x_center, -y_center)

        if "rotated" in options:
            x_center = rect.width() * 0.5
            y_center = rect.height() * 0.5
            painter.translate(x_center, y_center)
            painter.rotate(options["rotated"])
            painter.translate(-x_center, -y_center)

        painter.setOpacity(options.get("opacity", 1.0))

        painter.drawText(rect, int(constants.ALIGN_CENTER), char)
        painter.restore()


class FontError(Exception):
    """Exception for font errors."""


class CharIconEngine(gui.IconEngine):
    """Specialization of QtGui.QIconEngine used to draw font-based icons."""

    def __init__(self, iconic: QtCore.QObject, painter, options):
        super().__init__()
        self.iconic = iconic
        self.painter = painter
        self.options = options

    def paint(self, painter: QtGui.QPainter, rect: QtCore.QRect, mode, state):
        self.painter.paint(self.iconic, painter, rect, mode, state, self.options)

    def pixmap(self, size, mode, state) -> QtGui.QPixmap:
        pm = QtGui.QPixmap(size)
        pm.fill(QtCore.Qt.transparent)
        self.paint(gui.Painter(pm), core.Rect(ZERO_COORD, size), mode, state)
        return pm


class IconicFont(core.Object):
    """Main class for managing iconic fonts."""

    def __init__(self, *args):
        """Initialize IconicFont.

        Parameters
        ----------
        ``*args``: tuples
            Each positional argument is a tuple of 3 or 4 values:
            - The prefix string to be used when accessing a given font set,
            - The ttf font filename,
            - The json charmap filename,
            - Optionally, the directory containing these files. When not
              provided, the files will be looked for in ``./fonts/``.
        """
        super().__init__()
        self.painter = CharIconPainter()
        self.painters: Dict[str, CharIconPainter] = {}
        self.fontname = {}
        self.fontids = {}
        self.charmap = {}
        self.icon_cache = {}
        for fargs in args:
            self.load_font(*fargs)

    def load_font(
        self,
        prefix: str,
        ttf_filename: str,
        charmap_filename: str,
        directory: Optional[pathlib.Path] = None,
    ):
        """Load a font file and the associated charmap.

        If ``directory`` is None, the files will be looked for in ``./fonts/``.

        Parameters
        ----------
        prefix: str
            Prefix string to be used when accessing a given font set
        ttf_filename: str
            Ttf font filename
        charmap_filename: str
            Charmap filename
        directory: str or None, optional
            Directory for font and charmap files
        """
        if directory is None:
            directory = pathlib.Path(__file__).parent / "fonts"

        # Load font
        if gui.app() is None:
            return
        id_ = gui.FontDatabase.add_font(directory / ttf_filename)
        loaded_font_families = gui.FontDatabase.applicationFontFamilies(id_)

        if not loaded_font_families:
            raise FontError(
                f"Font at '{directory / ttf_filename}' appears to be empty. "
                "If you are on Windows 10, please read "
                "https://support.microsoft.com/"
                "en-us/kb/3053676 "
                "to know how to prevent Windows from blocking "
                "the fonts that come with QtAwesome."
            )
        self.fontids[prefix] = id_
        self.fontname[prefix] = loaded_font_families[0]

        def hook(obj: dict) -> dict:
            result = {}
            for key in obj:
                try:
                    result[key] = chr(int(obj[key], 16))
                except ValueError:
                    if int(obj[key], 16) > 0xFFFF:
                        # ignoring unsupported code in Python 2.7 32bit Windows
                        # ValueError: chr() arg not in range(0x10000)
                        pass
                    else:
                        raise FontError("Failed to load character " f"{key}:{obj[key]}")
            return result

        with (directory / charmap_filename).open("r") as codes:
            self.charmap[prefix] = json.load(codes, object_hook=hook)

        # Verify that vendorized fonts are not corrupt
        if SYSTEM_FONTS:
            return
        ttf_hash = MD5_HASHES.get(ttf_filename, None)
        if ttf_hash is None:
            return
        hasher = hashlib.md5()
        content = (directory / ttf_filename).read_bytes()
        hasher.update(content)
        ttf_calculated_hash_code = hasher.hexdigest()
        if ttf_calculated_hash_code != ttf_hash:
            raise FontError(f"Font is corrupt at: '{directory / ttf_filename}'")

    def icon(self, *names, **kwargs) -> QtGui.QIcon:
        """Return a QtGui.QIcon object corresponding to the provided icon name."""
        cache_key = f"{names}{kwargs}"
        if cache_key in self.icon_cache:
            return self.icon_cache[cache_key]
        options_list = kwargs.pop("options", [{}] * len(names))
        if len(options_list) != len(names):
            raise Exception(f'"options" must be a list of size {len(names)}')
        if gui.app() is None:
            warnings.warn("You need to have a running QGuiApplication.")
            return QtGui.QIcon()
        parsed_options = [
            self._parse_options(opt, kwargs, name)
            for opt, name in zip(options_list, names)
        ]
        icon = self._icon_by_painter(self.painter, parsed_options)
        self.icon_cache[cache_key] = icon
        return icon

    def _parse_options(
        self, specific_options: dict, general_options: dict, name: str
    ) -> Dict[str, Any]:
        options = dict(_default_options, **general_options)
        options.update(specific_options)

        # Handle icons for modes (Active, Disabled, Selected, Normal)
        # and states (On, Off)
        icon_kw = [
            "char",
            "on",
            "off",
            "active",
            "selected",
            "disabled",
            "on_active",
            "on_selected",
            "on_disabled",
            "off_active",
            "off_selected",
            "off_disabled",
        ]
        char = options.get("char", name)
        on = options.get("on", char)
        off = options.get("off", char)
        active = options.get("active", on)
        selected = options.get("selected", active)
        disabled = options.get("disabled", char)
        on_active = options.get("on_active", active)
        on_selected = options.get("on_selected", selected)
        on_disabled = options.get("on_disabled", disabled)
        off_active = options.get("off_active", active)
        off_selected = options.get("off_selected", selected)
        off_disabled = options.get("off_disabled", disabled)

        icon_dict = {
            "char": char,
            "on": on,
            "off": off,
            "active": active,
            "selected": selected,
            "disabled": disabled,
            "on_active": on_active,
            "on_selected": on_selected,
            "on_disabled": on_disabled,
            "off_active": off_active,
            "off_selected": off_selected,
            "off_disabled": off_disabled,
        }
        names = [icon_dict.get(kw, name) for kw in icon_kw]
        prefix, chars = self._get_prefix_chars(names)
        options.update(dict(zip(*(icon_kw, chars))))
        options.update({"prefix": prefix})

        # Handle colors for modes (Active, Disabled, Selected, Normal)
        # and states (On, Off)
        color = options.get("color")
        options.setdefault("color_on", color)
        options.setdefault("color_active", options["color_on"])
        options.setdefault("color_selected", options["color_active"])
        options.setdefault("color_on_active", options["color_active"])
        options.setdefault("color_on_selected", options["color_selected"])
        options.setdefault("color_on_disabled", options["color_disabled"])
        options.setdefault("color_off", color)
        options.setdefault("color_off_active", options["color_active"])
        options.setdefault("color_off_selected", options["color_selected"])
        options.setdefault("color_off_disabled", options["color_disabled"])

        return options

    def _get_prefix_chars(self, names: Iterable[str]) -> Tuple[str, List[str]]:
        chars = []
        for name in names:
            if "." not in name:
                raise Exception("Invalid icon name")
            prefix, n = name.split(".")
            if prefix not in self.charmap:
                raise Exception(f'Invalid font prefix "{prefix}"')
            if n not in self.charmap[prefix]:
                raise Exception(f'Invalid icon name "{n}" in font "{prefix}"')
            chars.append(self.charmap[prefix][n])
        return prefix, chars

    def font(self, prefix: str, size: float) -> gui.Font:
        """Return a gui.Font corresponding to the given prefix and size."""
        font = gui.Font(self.fontname[prefix])
        font.setPixelSize(round(size))
        if prefix[-1] == "s":  # solid style
            font.setStyleName("Solid")
        return font

    def set_custom_icon(self, name: str, painter: CharIconPainter):
        """Associate a user-provided CharIconPainter to an icon name.

        The custom icon can later be addressed by calling
        icon('custom.NAME') where NAME is the provided name for that icon.

        Parameters
        ----------
        name: str
            name of the custom icon
        painter: CharIconPainter
            The icon painter, implementing
            ``paint(self, iconic, painter, rect, mode, state, options)``
        """
        self.painters[name] = painter

    def _custom_icon(self, name: str, **kwargs) -> QtGui.QIcon:
        """Return the custom icon corresponding to the given name."""
        if name not in self.painters:
            return QtGui.QIcon()
        return self._icon_by_painter(
            self.painters[name], dict(_default_options, **kwargs)
        )

    def _icon_by_painter(self, painter: CharIconPainter, options) -> QtGui.QIcon:
        """Return the icon corresponding to the given painter."""
        engine = CharIconEngine(self, painter, options)
        return QtGui.QIcon(engine)



from typing import List

from qtpy import QtLocation

from prettyqt import core, location
from prettyqt.utils import bidict


QGeoServiceProvider = QtLocation.QGeoServiceProvider

ERROR = bidict(
    none=QGeoServiceProvider.NoError,
    not_supported=QGeoServiceProvider.NotSupportedError,
    unknown_parameter=QGeoServiceProvider.UnknownParameterError,
    missing_required_parameter=QGeoServiceProvider.MissingRequiredParameterError,
    connection=QGeoServiceProvider.ConnectionError,
    failed_to_load=QGeoServiceProvider.LoaderError,
)

GEOCODING_FEATURES = bidict(
    none=QtLocation.QGeoServiceProvider.NoGeocodingFeatures,
    online=QtLocation.QGeoServiceProvider.OnlineGeocodingFeature,
    offline=QtLocation.QGeoServiceProvider.OfflineGeocodingFeature,
    reverse=QtLocation.QGeoServiceProvider.ReverseGeocodingFeature,
    localized=QtLocation.QGeoServiceProvider.LocalizedGeocodingFeature,
    # any=QtLocation.QGeoServiceProvider.AnyGeocodingFeatures,
)

MAPPING_FEATURES = bidict(
    none=QtLocation.QGeoServiceProvider.NoMappingFeatures,
    online=QtLocation.QGeoServiceProvider.OnlineMappingFeature,
    offline=QtLocation.QGeoServiceProvider.OfflineMappingFeature,
    localized=QtLocation.QGeoServiceProvider.LocalizedMappingFeature,
    # any=QtLocation.QGeoServiceProvider.AnyMappingFeatures,
)

NAVIGATION_FEATURES = bidict(
    none=QtLocation.QGeoServiceProvider.NoNavigationFeatures,
    online=QtLocation.QGeoServiceProvider.OnlineNavigationFeature,
    offline=QtLocation.QGeoServiceProvider.OfflineNavigationFeature,
    # any=QtLocation.QGeoServiceProvider.AnyNavigationFeatures,
)

PLACES_FEATURES = bidict(
    none=QtLocation.QGeoServiceProvider.NoPlacesFeatures,
    online_places=QtLocation.QGeoServiceProvider.OnlinePlacesFeature,
    offline_places=QtLocation.QGeoServiceProvider.OfflinePlacesFeature,
    save_place=QtLocation.QGeoServiceProvider.SavePlaceFeature,
    remove_place=QtLocation.QGeoServiceProvider.RemovePlaceFeature,
    save_category=QtLocation.QGeoServiceProvider.SaveCategoryFeature,
    remove_category=QtLocation.QGeoServiceProvider.RemoveCategoryFeature,
    place_recommendations=QtLocation.QGeoServiceProvider.PlaceRecommendationsFeature,
    search_suggestions=QtLocation.QGeoServiceProvider.SearchSuggestionsFeature,
    localized_places=QtLocation.QGeoServiceProvider.LocalizedPlacesFeature,
    notifications=QtLocation.QGeoServiceProvider.NotificationsFeature,
    place_matching=QtLocation.QGeoServiceProvider.PlaceMatchingFeature,
    # any=QtLocation.QGeoServiceProvider.AnyPlacesFeatures,
)

ROUTING_FEATURES = bidict(
    none=QtLocation.QGeoServiceProvider.NoRoutingFeatures,
    online=QtLocation.QGeoServiceProvider.OnlineRoutingFeature,
    offline=QtLocation.QGeoServiceProvider.OfflineRoutingFeature,
    localized=QtLocation.QGeoServiceProvider.LocalizedRoutingFeature,
    route_updates=QtLocation.QGeoServiceProvider.RouteUpdatesFeature,
    alternative_routes=QtLocation.QGeoServiceProvider.AlternativeRoutesFeature,
    exclude_areas=QtLocation.QGeoServiceProvider.ExcludeAreasRoutingFeature,
    # any=QtLocation.QGeoServiceProvider.AnyRoutingFeatures,
)


QtLocation.QGeoServiceProvider.__bases__ = (core.Object,)


class GeoServiceProvider(QtLocation.QGeoServiceProvider):
    def get_error(self) -> str:
        return ERROR.inverse[self.error()]

    def get_geocoding_error(self) -> str:
        return ERROR.inverse[self.geocodingError()]

    def get_geocoding_features(self) -> List[str]:
        return [k for k, v in GEOCODING_FEATURES.items() if v & self.geocodingFeatures()]

    def get_mapping_error(self) -> str:
        return ERROR.inverse[self.mappingError()]

    def get_mapping_features(self) -> List[str]:
        return [k for k, v in MAPPING_FEATURES.items() if v & self.mappingFeatures()]

    def get_navigation_error(self) -> str:
        return ERROR.inverse[self.navigationError()]

    def get_navigation_features(self) -> List[str]:
        return [
            k for k, v in NAVIGATION_FEATURES.items() if v & self.navigationFeatures()
        ]

    def get_places_error(self) -> str:
        return ERROR.inverse[self.placesError()]

    def get_places_features(self) -> List[str]:
        return [k for k, v in PLACES_FEATURES.items() if v & self.placesFeatures()]

    def get_routing_error(self) -> str:
        return ERROR.inverse[self.routingError()]

    def get_routing_features(self) -> List[str]:
        return [k for k, v in ROUTING_FEATURES.items() if v & self.routingFeatures()]

    def get_geocoding_manager(self) -> location.GeoCodingManager:
        return location.GeoCodingManager(self.geocodingManager())

    def get_routing_manager(self) -> location.GeoRoutingManager:
        return location.GeoRoutingManager(self.routingManager())

    def get_place_manager(self) -> location.PlaceManager:
        return location.PlaceManager(self.placeManager())



from qtpy import QtCore


class PersistentModelIndex(QtCore.QPersistentModelIndex):
    def __bool__(self):
        return self.isValid()

    def __getitem__(self, flag: int):
        return self.data(flag)


if __name__ == "__main__":
    index = PersistentModelIndex()
    print(bool(index))



import pathlib
from typing import Union

from qtpy import QtCore, QtGui

from prettyqt import core


class DesktopServices(QtGui.QDesktopServices):
    @classmethod
    def open_url(cls, location: Union[str, pathlib.Path, QtCore.QUrl]) -> bool:
        return cls.openUrl(core.Url.from_user_input(str(location)))



from typing import Callable, Optional

from qtpy import QtWidgets

from prettyqt import core, widgets


QtWidgets.QPushButton.__bases__ = (widgets.AbstractButton,)


class PushButton(QtWidgets.QPushButton):

    value_changed = core.Signal(bool)

    def __init__(
        self,
        label: Optional[str] = None,
        parent: Optional[QtWidgets.QWidget] = None,
        callback: Optional[Callable] = None,
    ):
        if label is None:
            label = ""
        super().__init__(label, parent)
        if callback:
            self.clicked.connect(callback)
        self.toggled.connect(self.value_changed)

    def get_value(self) -> bool:
        return self.isChecked()

    def set_value(self, value: bool):
        self.setChecked(value)

    @property
    def is_on(self) -> bool:
        return self.isChecked()

    @is_on.setter
    def is_on(self, state: bool):
        self.setChecked(state)


if __name__ == "__main__":
    app = widgets.app()
    widget = PushButton("This is a test")
    widget.show()
    app.main_loop()



import pathlib
from typing import Union

from qtpy import PYQT5, PYSIDE2


if PYQT5:
    from PyQt5 import QtWinExtras  # type: ignore
elif PYSIDE2:
    from PySide2 import QtWinExtras

from prettyqt import iconprovider
from prettyqt.utils import bidict


TYPES = bidict(
    destination=QtWinExtras.QWinJumpListItem.Destination,
    link=QtWinExtras.QWinJumpListItem.Link,
    separator=QtWinExtras.QWinJumpListItem.Separator,
)


class WinJumpListItem(QtWinExtras.QWinJumpListItem):
    def __init__(self, typ: Union[QtWinExtras.QWinJumpListItem.Type, str]) -> None:
        if isinstance(typ, QtWinExtras.QWinJumpListItem.Type):
            param = typ
        else:
            param = TYPES[typ]
        super().__init__(param)

    def set_title(self, title: str) -> None:
        self.setTitle(title)

    def set_icon(self, icon: iconprovider.IconType) -> None:
        icon = iconprovider.get_icon(icon)
        self.setIcon(icon)

    def set_file_path(self, path: Union[str, pathlib.Path]) -> None:
        self.setFilePath(str(path))

    def get_file_path(self) -> pathlib.Path:
        return pathlib.Path(self.filePath())

    def set_working_directory(self, path: Union[str, pathlib.Path]) -> None:
        self.setWorkingDirectory(str(path))

    def get_working_directory(self) -> pathlib.Path:
        return pathlib.Path(self.workingDirectory())

    def set_type(self, typ: str) -> None:
        self.setType(TYPES[typ])

    def get_type(self) -> str:
        return TYPES.inverse[self.type()]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    pass
    app.main_loop()



from qtpy.QtCharts import QtCharts

from prettyqt import core, gui


QtCharts.QBarSet.__bases__ = (core.Object,)


class BarSet(QtCharts.QBarSet):
    def __delitem__(self, index: int):
        self.remove(index)

    def __getitem__(self, label: int) -> float:
        return self.at(label)

    def __setitem__(self, index: int, value: float):
        self.replace(index, value)

    def __repr__(self):
        return f"{type(self).__name__}({self.label()!r})"

    def get_pen(self) -> gui.Pen:
        return gui.Pen(self.pen())

    def get_brush(self) -> gui.Brush:
        return gui.Brush(self.brush())


if __name__ == "__main__":
    barset = BarSet("test")
    print(repr(barset))



from typing import Literal

from qtpy import QtCore

from prettyqt import core
from prettyqt.utils import InvalidParamError, bidict


TRANSITION_TYPE = bidict(
    exclusive=QtCore.QAbstractTransition.ExternalTransition,
    parallel=QtCore.QAbstractTransition.InternalTransition,
)

TransitionTypeStr = Literal["exclusive", "parallel"]

QtCore.QAbstractTransition.__bases__ = (core.Object,)


class AbstractTransition(QtCore.QAbstractTransition):
    def set_transition_type(self, typ: TransitionTypeStr):
        """Set transition type.

        Args:
            typ: transition type to use

        Raises:
            InvalidParamError: transition type does not exist
        """
        if typ not in TRANSITION_TYPE:
            raise InvalidParamError(typ, TRANSITION_TYPE)
        self.setTransitionType(TRANSITION_TYPE[typ])

    def get_transition_type(self) -> TransitionTypeStr:
        """Return current transition type.

        Returns:
            transition type
        """
        return TRANSITION_TYPE.inverse[self.transitionType()]



from qtpy import QtGui

from prettyqt import gui
from prettyqt.utils import InvalidParamError, bidict


STYLES = bidict(
    disc=QtGui.QTextListFormat.ListDisc,
    circle=QtGui.QTextListFormat.ListCircle,
    square=QtGui.QTextListFormat.ListSquare,
    decimal=QtGui.QTextListFormat.ListDecimal,
    lower_alpha=QtGui.QTextListFormat.ListLowerAlpha,
    upper_alpha=QtGui.QTextListFormat.ListUpperAlpha,
    lower_roman=QtGui.QTextListFormat.ListLowerRoman,
    upper_roman=QtGui.QTextListFormat.ListUpperRoman,
)


QtGui.QTextListFormat.__bases__ = (gui.TextFormat,)


class TextListFormat(QtGui.QTextListFormat):
    def set_style(self, style: str):
        """Set the style.

        Valid values are "disc", "circle", square", "decimal", "lower_alpha",
                         "upper_alpha", "lower_roman", upper_roman"

        Args:
            style: style

        Raises:
            InvalidParamError: invalid style
        """
        if style not in STYLES:
            raise InvalidParamError(style, STYLES)
        self.setStyle(STYLES[style])

    def get_style(self) -> str:
        """Get current style.

        Possible values are "disc", "circle", square", "decimal", "lower_alpha",
                            "upper_alpha", "lower_roman", upper_roman"

        Returns:
            current style
        """
        return STYLES.inverse[self.style()]



from qtpy import QtLocation

from prettyqt import core


QtLocation.QGeoCodingManager.__bases__ = (core.Object,)


class GeoCodingManager:
    def __init__(self, item: QtLocation.QGeoCodingManager):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_locale(self) -> core.Locale:
        return core.Locale(self.item.locale())



from typing import List, Sequence, Union

from qtpy import QtCore, QtHelp

from prettyqt import core


class HelpFilterData(QtHelp.QHelpFilterData):
    def set_versions(self, versions: Sequence[Union[QtCore.QVersionNumber, str]]):
        versions = [core.VersionNumber(i) for i in versions]
        self.setVersions(versions)

    def get_versions(self) -> List[core.VersionNumber]:
        return [core.VersionNumber(i) for i in self.versions()]



from __future__ import annotations

from typing import Literal

from qtpy import PYQT5, PYSIDE2, QtCore


if PYQT5:
    from PyQt5 import QtWebEngineCore  # type: ignore
elif PYSIDE2:
    from PySide2 import QtWebEngineCore

from prettyqt.utils import InvalidParamError, bidict


QWebEngineUrlScheme = QtWebEngineCore.QWebEngineUrlScheme

SYNTAX = bidict(
    host_port_user_info=QWebEngineUrlScheme.Syntax.HostPortAndUserInformation,
    host_port=QWebEngineUrlScheme.Syntax.HostAndPort,
    host=QWebEngineUrlScheme.Syntax.Host,
    path=QWebEngineUrlScheme.Syntax.Path,
)

SyntaxStr = Literal["host_port_user_info", "host_port", "host", "path"]

FLAGS = bidict(
    secure_scheme=QWebEngineUrlScheme.SecureScheme,
    local_scheme=QWebEngineUrlScheme.LocalScheme,
    local_access_allowed=QWebEngineUrlScheme.LocalAccessAllowed,
    no_access_allowed=QWebEngineUrlScheme.NoAccessAllowed,
    service_workers_allowed=QWebEngineUrlScheme.ServiceWorkersAllowed,
    view_source_allowed=QWebEngineUrlScheme.ViewSourceAllowed,
    content_security_policy_ignored=QWebEngineUrlScheme.ContentSecurityPolicyIgnored,
    cors_enabled=QWebEngineUrlScheme.CorsEnabled,
)


class WebEngineUrlScheme(QtWebEngineCore.QWebEngineUrlScheme):
    def get_name(self) -> str:
        return bytes(self.name()).decode()

    @classmethod
    def get_scheme_by_name(cls, name: str) -> WebEngineUrlScheme:
        scheme = cls.schemeByName(QtCore.QByteArray(name.encode()))
        return cls(scheme)

    def set_name(self, name: str):
        self.setName(QtCore.QByteArray(name.encode()))

    def set_syntax(self, syntax: SyntaxStr):
        """Set syntax.

        Args:
            syntax: syntax to use

        Raises:
            InvalidParamError: syntax does not exist
        """
        if syntax not in SYNTAX:
            raise InvalidParamError(syntax, SYNTAX)
        self.setSyntax(SYNTAX[syntax])

    def get_syntax(self) -> SyntaxStr:
        """Return syntax.

        Returns:
            syntax
        """
        return SYNTAX.inverse[self.syntax()]


if __name__ == "__main__":
    item = WebEngineUrlScheme()



import pathlib
from typing import Optional, Union

from qtpy import QtCore, QtGui, QtWidgets

from prettyqt import constants, gui, widgets
from prettyqt.utils import colors


QtWidgets.QSplashScreen.__bases__ = (widgets.Widget,)


class SplashScreen(QtWidgets.QSplashScreen):
    def __init__(
        self, path: Union[pathlib.Path, str, QtGui.QPixmap], width: Optional[int] = None
    ):
        pix = gui.Pixmap(str(path)) if not isinstance(path, QtGui.QPixmap) else path
        if width:
            pix = pix.scaledToWidth(width)
        super().__init__(pix)
        self.setWindowFlags(
            QtCore.Qt.WindowStaysOnTopHint | QtCore.Qt.FramelessWindowHint
        )
        self.setEnabled(False)

    def __enter__(self):
        self.show()
        return self

    def __exit__(self, typ, value, traceback):
        self.hide()

    def set_text(
        self,
        text: str,
        color: colors.ColorType = "black",
        h_align: constants.HorizontalAlignmentStr = "center",
        v_align: constants.VerticalAlignmentStr = "bottom",
    ):
        self.showMessage(
            text,
            color=gui.Color(color),
            alignment=constants.H_ALIGNMENT[h_align] | constants.V_ALIGNMENT[v_align],
        )



from qtpy import QtCore


class Rect(QtCore.QRect):
    def __repr__(self):
        return (
            f"{type(self).__name__}({self.x()}, {self.y()}, "
            f"{self.width()}, {self.height()})"
        )

    def __reduce__(self):
        return self.__class__, (self.x(), self.y(), self.width(), self.height())


if __name__ == "__main__":
    rect = Rect(0, 2, 5, 5)
    print(repr(rect))



# see https://github.com/ITVRoC/SeekurJr/blob/master/seekur_12.04/packages/
# multimaster_fkie/node_manager_fkie/src/node_manager_fkie/yaml_highlighter.py


from prettyqt import core, gui, syntaxhighlighters


COMMENT_START = core.RegularExpression("#")
COMMENT_END = core.RegularExpression("\n|\r")  # Unused?
COMMENT_FORMAT = gui.TextCharFormat()
COMMENT_FORMAT.setFontItalic(True)
COMMENT_FORMAT.set_foreground_color("darkgray")


class Rule(syntaxhighlighters.HighlightRule):
    minimal = True


class Bool(Rule):
    regex = [r"\btrue\b", r"\bfalse\b"]
    color = "blue"
    bold = True


class Decimal(Rule):
    regex = r"\d+"
    color = "darkMagenta"


class Rule2(Rule):
    regex = r"^\s*[_.\w]*\s*:"
    color = "blue"


class Rule3(Rule):
    regex = r":\s*:[_\.\w]*$|:\s*\@[_\.\w]*$"
    color = "blue"


class ListMember(Rule):
    regex = r"^\s*-"
    color = "red"
    bold = True


class DocumentStart(Rule):
    regex = r"^---$"
    color = "red"
    bold = True


class Brackets(Rule):
    regex = r"[\[\]\{\}\,]"
    color = "darkgreen"
    bold = True


class Rule7(Rule):
    regex = r"\".*\"|\'.*\'"
    color = "darkorange"


class Rule8(Rule):
    regex = r"\$\(.*\)"
    color = "orange"


class DocType(Rule):
    regex = r"<!DOCTYPE.*>"
    color = "lightgray"


class Xml(Rule):
    regex = r"<\?xml.*\?>"
    color = "lightgray"


class YamlHighlighter(gui.SyntaxHighlighter):
    RULES = Rule.__subclasses__()

    def highlightBlock(self, text: str):
        super().highlightBlock(text)
        self.setCurrentBlockState(0)
        start_index = 0
        if self.previousBlockState() != 1:
            start_index = COMMENT_START.match(text).capturedStart()
            if start_index >= 0:
                comment_len = len(text) - start_index
                self.setFormat(start_index, comment_len, COMMENT_FORMAT)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    editor = widgets.PlainTextEdit()
    highlighter = YamlHighlighter(editor.document())
    editor.show()
    app.main_loop()



from typing import Optional

from qtpy import QtWidgets

from prettyqt import constants, widgets


QtWidgets.QGroupBox.__bases__ = (widgets.Widget,)


class GroupBox(QtWidgets.QGroupBox):
    """GroupBox widget.

    A group box provides a frame, a title on top, a keyboard shortcut,
    and displays various other widgets inside itself.
    The keyboard shortcut moves keyboard focus to one of the group box's child widgets.
    """

    def __init__(
        self,
        title: str = "",
        checkable: bool = False,
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        super().__init__(title, parent)
        self.setCheckable(checkable)

    def __repr__(self):
        return f"{type(self).__name__}({self.title()!r})"

    def serialize_fields(self):
        return dict(
            checkable=self.isCheckable(),
            checked=self.isChecked(),
            layout=self.layout(),
            flat=self.isFlat(),
            # alignment=self.alignment(),
            title=self.title(),
        )

    def __setstate__(self, state):
        self.setTitle(state["title"])
        self.set_layout(state["layout"])
        self.setCheckable(state["checkable"])
        self.setChecked(state.get("checked", False))
        self.setFlat(state["flat"])
        self.setToolTip(state.get("tool_tip", ""))
        # self.setAlignment(state["alignment"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def set_title(self, title: str):
        self.setTitle(title)

    def set_alignment(self, alignment):
        self.setAlignment(constants.H_ALIGNMENT[alignment])

    def set_enabled(self, state):
        for widget in self.layout():
            widget.setEnabled(state)


if __name__ == "__main__":
    app = widgets.app()
    widget = GroupBox()
    ly = widgets.BoxLayout()
    ly += widgets.RadioButton("test")
    widget.set_layout(ly)
    widget.show()
    app.main_loop()



from typing import Literal

from qtpy import QtGui

from prettyqt import core
from prettyqt.utils import bidict


REGION_TYPE = bidict(
    rectangle=QtGui.QRegion.Rectangle,
    ellipse=QtGui.QRegion.Ellipse,
)

RegionTypeStr = Literal["rectangle", "ellipse"]


class Region(QtGui.QRegion):
    def __repr__(self):
        return f"{type(self).__name__}({self.get_bounding_rect()}, {self.get_shape()!r})"

    def __getstate__(self):
        return bytes(self)

    def __setstate__(self, ba):
        core.DataStream.write_bytearray(ba, self)

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __bytes__(self):
        ba = core.DataStream.create_bytearray(self)
        return bytes(ba)

    def get_bounding_rect(self) -> core.Rect:
        return core.Rect(self.boundingRect())

    def get_shape(self) -> RegionTypeStr:  # workaround for not being able to get shape
        return "rectangle" if self == Region(self.get_bounding_rect()) else "ellipse"


if __name__ == "__main__":
    region = Region(0, 0, 10, 10)
    print(repr(region))



from __future__ import annotations

from typing import Iterator, Literal, Tuple, Union

from qtpy import QtCore, QtGui

from prettyqt import constants, core
from prettyqt.utils import InvalidParamError, bidict


ELEMENT_TYPES = bidict(
    move_to_element=QtGui.QPainterPath.MoveToElement,
    line_to_element=QtGui.QPainterPath.LineToElement,
    curve_to_element=QtGui.QPainterPath.CurveToElement,
    curve_to_data_element=QtGui.QPainterPath.CurveToDataElement,
)

ElementTypeStr = Literal[
    "move_to_element", "line_to_element", "curve_to_element", "curve_to_data_element"
]


class PainterPath(QtGui.QPainterPath):
    def serialize_fields(self):
        return dict(fill_rule=self.get_fill_rule(), elements=list(self))

    def __len__(self):
        return self.elementCount()

    def __getitem__(self, index: int) -> QtGui.QPainterPath.Element:
        return self.elementAt(index)

    def __iter__(self) -> Iterator[QtGui.QPainterPath.Element]:
        return iter(self.elementAt(i) for i in range(self.elementCount()))

    def __setitem__(self, index: int, value: Tuple[int, int]):
        self.setElementPositionAt(index, *value)

    def __bool__(self):
        return not self.isEmpty()

    def __contains__(self, item: Union[QtCore.QPointF, QtCore.QRectF, PainterPath]):
        return self.contains(item)

    def add_rect(self, rect: Union[QtCore.QRectF, QtCore.QRect]):
        if isinstance(rect, QtCore.QRect):
            rect = QtCore.QRectF(rect)
        self.addRect(rect)

    def set_fill_rule(self, rule: constants.FillRuleStr):
        """Set fill rule.

        Args:
            rule: fill rule to use

        Raises:
            InvalidParamError: fill rule does not exist
        """
        if rule not in constants.FILL_RULE:
            raise InvalidParamError(rule, constants.FILL_RULE)
        self.setFillRule(constants.FILL_RULE[rule])

    def get_fill_rule(self) -> constants.FillRuleStr:
        """Return current fill rule.

        Returns:
            fill rule
        """
        return constants.FILL_RULE.inverse[self.fillRule()]

    def get_bounding_rect(self) -> core.RectF:
        return core.RectF(self.boundingRect())


if __name__ == "__main__":
    p = PainterPath(QtCore.QPoint(1, 1))
    print(type(p[0]))
